# handle the metal: namespace

metal_ns = 'http://xml.zope.org/namespaces/metal'

from pax.backwards_compatibility import *
from pax.paxtransform import AttributeHandler
from xml.sax.saxutils import escape as html_quote
from pax import XML, Element
import pax
try:
    from tal_handler import tal_ns
except:
    tal_ns = []

def text_or_structure(value):
    return split_spec(value, 'text', 'structure')

def split_spec(value, *names):
    for name in names:
        if value.startswith(name + ' '):
            return name, value[len(name)+1:]
    return names[0], value

def remove_attrs_till(name, element):
    try:
        attrs = element.attributes[metal_ns]
    except KeyError:
        return # no metal attributes
    names = metal_handler.order
    for n in names:
        try:
            del attrs[n]
        except KeyError:
            pass
        if n == name:
            break

default = []

def pax_fish(element, attrname, deep=0):
    found = {}
    if hasattr(element, 'attributes'):
        if element.attributes.has_key(metal_ns):
            metal = element.attributes[metal_ns]
            if metal.has_key(attrname):
                found[metal[attrname]] = element.copy()
    if deep or not found:
        try:
            for child in element:
                found.update(pax_fish(child, attrname, deep))
        except:
            pass
    return found

def fish_slots(element):
    if hasattr(element, 'metal_slots'):
        return element.metal_slots.copy()
    return pax_fish(element, 'fill-slot')

def fill_slots(macro, slots):
    for pos, child in enumerate(macro):
        if hasattr(child, 'attributes'):
            if (child.attributes.has_key(metal_ns) and
                child.attributes[metal_ns].has_key('define-slot')):
                name = child.attributes[metal_ns]['define-slot']
                macro[pos] = slots.get(name, child)
            else:
                fill_slots(child, slots)
    return macro

class MetalError(Exception):
    pass


# the MeTAL interpreter

def handle_use_macro(element, context, value):
    macro = context.engine.evaluateMacro(value)
    if not isinstance(macro, Element):
        raise MetalError, ('trying to use a PageTemplate macro in an AltPageTemplate: ' +
                           value)
    if hasattr(context.engine, 'getDefault') and macro is context.engine.getDefault():
        return element
    context.metal_slot_stack[-1] = fish_slots(element)
    result = macro.copy()
    metal = result.attributes[metal_ns]
    del metal['define-macro']
    metal['use-macro'] = value
    return result

def handle_define_slot(element, context, value):
    layers = [layer for layer in context.metal_slot_stack if layer]
    layers.reverse()
    for layer in layers:
        if layer.has_key(value):
            slot = layer[value]
            del layer[value]
            return slot.copy()
    return element


def metal_initialize(context):
    context.metal_slot_stack = []

def metal_start(element, context):
    context.metal_slot_stack.append(None)
    return element

def metal_finish(element, context):
    context.metal_slot_stack.pop()
    if hasattr(element, 'attributes') and element.attributes.has_key(metal_ns):
        # have to beware, element may be an unicode
        del element.attributes[metal_ns]
    if hasattr(element, 'nsdecls'):
        for dprefix, duri in element.nsdecls.items():
            if duri == metal_ns:
                del element.nsdecls[dprefix]
    if getattr(element, 'ns', None) == metal_ns:
        if getattr(element, 'attributes', {}).has_key(tal_ns):
            element.ns = tal_ns
        else:
            return element.children
    return element


metal_handler = AttributeHandler('metal', metal_ns)
metal_handler.add_handler('use-macro', handle_use_macro)
metal_handler.add_handler('define-slot', handle_define_slot)
metal_handler.preprocess = metal_start
metal_handler.postprocess = metal_finish
metal_handler.initialize = metal_initialize

# the MeTAL compiler

def compile_use_macro(element, context, value):
    slots = {}
    element.metal_slots = slots
    context.metal_slot_stack[-1] = slots
    return element

def compile_define_macro(element, context, value):
    if not hasattr(context, 'metal_macros'):
        context.metal_macros = {}
    context.metal_macros[value] = element
    return element

def compile_fill_slot(element, context, value):
    layers = [layer for layer in context.metal_slot_stack if layer is not None]
    if not layers:
        raise MetalError, 'fill-slot outside use-macro'
    layers[-1][value] = element
    return element


def compile_metal_finish(element, context):
    context.metal_slot_stack.pop()
    return element


metal_compiler = AttributeHandler('metal', metal_ns)
metal_compiler.add_handler('define-macro', compile_define_macro)
metal_compiler.add_handler('use-macro', compile_use_macro)
metal_compiler.add_handler('fill-slot', compile_fill_slot)
metal_compiler.preprocess = metal_start
metal_compiler.postprocess = compile_metal_finish
metal_compiler.initialize = metal_initialize


__all__ = ('metal_handler', 'metal_compiler')



if __name__ == '__main__':
    metal_tr_engine = pax.paxtransform.Engine()
    metal_tr_engine.add_handler(metal_ns, metal_handler)
    metal_co_engine = pax.paxtransform.Engine()
    metal_co_engine.add_handler(metal_ns, metal_compiler)
    import sys, pax, pax.paxtransform
    import DummyEngine
    default = DummyEngine.Default

    def metal_compilefile(self, name):
        if name[:-5] == '.html':
            tree = pax.html2pax(open(name).read())
        else:
            tree = pax.xml2pax(open(name).read())
        tree = metal_co_engine.transform(tree, self)
        return tree

    def metal_getmacros(self, tree):
        return self.metal_macros

    DummyEngine.DummyEngine.compilefile = metal_compilefile
    DummyEngine.DummyEngine.getmacros = metal_getmacros

    class DumbContext:
        def __init__(self, tr_engine, macros=None):
            self.engine = DummyEngine.DummyEngine(macros)
            self.tr_engine = tr_engine
    if len(sys.argv) > 1:
        srcfile = open(sys.argv[1])
    else:
        srcfile = sys.stdin
        sys.stderr.write('Reading source from stdin\n')
    source = pax.xml2pax(srcfile.read())
    compile_context = DumbContext(metal_co_engine)
    source = metal_co_engine.transform(source, compile_context)
    macros = getattr(compile_context, 'metal_macros', {})
    print pax.XML(metal_tr_engine.transform(source, DumbContext(metal_tr_engine, macros)))
